`default_nettype none

module mos6581 (
  input wire clk,
  input wire clk_en,
  input wire n_reset,
  input wire [4:0] addr,
  input wire [7:0] din,       // TODO: inout
  input wire n_cs,
  input wire rw,
  output wire [15:0] audio_out
  );

  // CANAL 0
  // register reg s
  //
  reg [15:0]   freq0 = 16'h0000;
  reg [11:0]   pw0 = 12'h000;
  reg          noise0 = 1'b0, pulse0 = 1'b0, saw0 = 1'b0, triangle0 = 1'b0;
  reg          test0 = 1'b0, ring0 = 1'b0, sync0 = 1'b0, gate0 = 1'b0;
  reg [3:0]    atk0 = 1'b0, dcy0 = 1'b0, stn0 = 1'b0, rls0 = 1'b0;

  // internal state
  //
  wire [23:0]   acc0;
  wire [22:0]   lfsr0;
  wire          sync_out0;
  wire [7:0]    env_vol0;
  wire [11:0]   out0;

  // CANAL 1
  // register reg s
  //
  reg [15:0]   freq1 = 16'h0000;
  reg [11:0]   pw1 = 12'h000;
  reg          noise1 = 1'b0, pulse1 = 1'b0, saw1 = 1'b0, triangle1 = 1'b0;
  reg          test1, ring1 = 1'b0, sync1 = 1'b0, gate1 = 1'b0;
  reg [3:0]    atk1 = 1'b0, dcy1 = 1'b0, stn1 = 1'b0, rls1 = 1'b0;

  // internal state
  //
  wire [23:0]   acc1;
  wire [22:0]   lfsr1;
  wire          sync_out1;
  wire [7:0]    env_vol1;
  wire [11:0]   out1;

  // CANAL 2
  // register reg s
  //
  reg [15:0]   freq2 = 16'h0000;
  reg [11:0]   pw2 = 12'h000;
  reg          noise2 = 1'b0, pulse2 = 1'b0, saw2 = 1'b0, triangle2 = 1'b0;
  reg          test2, ring2 = 1'b0, sync2 = 1'b0, gate2 = 1'b0;
  reg [3:0]    atk2, dcy2 = 1'b0, stn2 = 1'b0, rls2 = 1'b0;

  // internal state
  //
  wire [23:0]   acc2;
  wire [22:0]   lfsr2;
  wire          sync_out2;
  wire [7:0]    env_vol2;
  wire [11:0]   out2;


  // Cosas del filtro
  //
  reg [10:0]   fc = 11'h000;
  reg [3:0]    res = 4'h0;
  reg [3:0]    filt = 4'h0;
  reg          off3 = 1'b0, hp = 1'b0, bp = 1'b0, lp = 1'b0;
  reg [3:0]    vol = 4'h0;


  // ---------- Memory interface ----------
  //
  always @(posedge clk) begin
    if (!n_reset) begin
      // TODO: reset all registers
    end
    else if (!n_cs && !rw) begin
      case(addr)
        // Voice 0
        //
        5'h00:   freq0[7:0]  <= din;
        5'h01:   freq0[15:8] <= din;
        5'h02:   pw0[7:0]    <= din;
        5'h03:   pw0[11:8]   <= din[3:0];
        5'h04:   { noise0, pulse0, saw0, triangle0,
                  test0,  ring0,  sync0, gate0 } <= din;
        5'h05:   { atk0, dcy0 } <= din;
        5'h06:   { stn0, rls0 } <= din;

        // Voice 1
        //
        5'h07:   freq1[7:0]  <= din;
        5'h08:   freq1[15:8] <= din;
        5'h09:   pw1[7:0]    <= din;
        5'h0A:   pw1[11:8]   <= din[3:0];
        5'h0B:   { noise1, pulse1, saw1,  triangle1,
                  test1,  ring1,  sync1, gate1 } <= din;
        5'h0C:   { atk1, dcy1 } <= din;
        5'h0D:   { stn1, rls1 } <= din;

        // Voice 2
        //
        5'h0E:   freq2[7:0]  <= din;
        5'h0F:   freq2[15:8] <= din;
        5'h10:   pw2[7:0]    <= din;
        5'h11:   pw2[11:8]   <= din[3:0];
        5'h12:   { noise2, pulse2, saw2,  triangle2,
                  test2,  ring2,  sync2, gate2 } <= din;
        5'h13:   { atk2, dcy2 } <= din;
        5'h14:   { stn2, rls2 } <= din;

        // Filter
        //
        5'h15:   fc[2:0]  <= din[2:0];
        5'h16:   fc[10:3] <= din;
        5'h17:   { res, filt } <= din;
        5'h18:   { off3, hp, bp, lp, vol } <= din;

      endcase
    end
  end

  sid_acc acum0 (
    .clk(clk),
    .clk_en(clk_en),
    .n_reset(n_reset),
    .freq(freq0),
    .lfsr(lfsr0),
    .test(test0),
    .sync(sync0),
    .sync_in(sync_out2),
    .sync_out(sync_out0),
    .acc (acc0)
  );

  sid_acc acum1 (
    .clk(clk),
    .clk_en(clk_en),
    .n_reset(n_reset),
    .freq(freq1),
    .lfsr(lfsr1),
    .test(test1),
    .sync(sync1),
    .sync_in(sync_out0),
    .sync_out(sync_out1),
    .acc (acc1)
  );

  sid_acc acum2 (
    .clk(clk),
    .clk_en(clk_en),
    .n_reset(n_reset),
    .freq(freq2),
    .lfsr(lfsr2),
    .test(test2),
    .sync(sync2),
    .sync_in(sync_out1),
    .sync_out(sync_out2),
    .acc (acc2)
  );

  sid_env env0 (
    .clk(clk),
    .clk_en(clk_en),
    .n_reset(n_reset),
    .gate(gate0),
    .atk(atk0),
    .dcy(dcy0),
    .stn(stn0),
    .rls(rls0),
    .vol(env_vol0)
  );

  sid_env env1 (
    .clk(clk),
    .clk_en(clk_en),
    .n_reset(n_reset),
    .gate(gate1),
    .atk(atk1),
    .dcy(dcy1),
    .stn(stn1),
    .rls(rls1),
    .vol(env_vol1)
  );

  sid_env env2 (
    .clk(clk),
    .clk_en(clk_en),
    .n_reset(n_reset),
    .gate(gate2),
    .atk(atk2),
    .dcy(dcy2),
    .stn(stn2),
    .rls(rls2),
    .vol(env_vol2)
  );

  sid_wave wave0 (
    .acc(acc0),
    .lfsr(lfsr0),
    .vol(env_vol0),
    .noise(noise0),
    .pulse(pulse0),
    .saw(saw0),
    .triangle(triangle0),
    .pw(pw0),
    .ring(ring0),
    .ring_in(acc2[23]),
    .out(out0)
  );

  sid_wave wave1 (
    .acc(acc1),
    .lfsr(lfsr1),
    .vol(env_vol1),
    .noise(noise1),
    .pulse(pulse1),
    .saw(saw1),
    .triangle(triangle1),
    .pw(pw1),
    .ring(ring1),
    .ring_in(acc0[23]),
    .out(out1)
  );

  sid_wave wave2 (
    .acc(acc2),
    .lfsr(lfsr2),
    .vol(env_vol2),
    .noise(noise2),
    .pulse(pulse2),
    .saw(saw2),
    .triangle(triangle2),
    .pw(pw2),
    .ring(ring2),
    .ring_in(acc1[23]),
    .out(out2)
  );

  sid_filter filtro (
    .clk(clk),
    .clk_en(clk_en),
    .n_reset(n_reset),
    .v0         ( out0   ),
    .v1         ( out1   ),
    .v2         ( out2   ),
    .reg_fc     ( fc     ),
    .reg_res    ( res    ),
    .reg_en     ( filt   ),
    .reg_off3   ( off3   ),
    .reg_hp     ( hp     ),
    .reg_bp     ( bp     ),
    .reg_lp     ( lp     ),
    .reg_vol    ( vol    ),
    .audio_out  ( audio_out )
  );

//  assign audio_out = out0 + out1 + out2;

endmodule
