`default_nettype none

// SID phase accumulator
//
module sid_acc (
  input wire clk,
  input wire clk_en,
  input wire n_reset,
  input wire [15:0] freq,
  input wire test,
  input wire sync,
  input wire sync_in,
  output reg [23:0] acc,
  output reg [22:0] lfsr,
  output reg sync_out
  );

  reg [23:0] acc_next;
  reg [22:0] lfsr_next;

  always @(posedge clk) begin
    if (!n_reset) begin
      acc <= 24'd0;
      lfsr <= 23'h7FFFFF;
    end
    else if (clk_en) begin
      acc <= acc_next;
      if (!acc[19] && acc_next[19])
        lfsr <= lfsr_next;
    end
  end

  always @* begin
    acc_next = acc + freq;
    lfsr_next = { lfsr[21:0], lfsr[17] ^ lfsr[22] };

    if (sync && sync_in)
      acc_next = 24'd0;

    if (test) begin
      acc_next = 24'd0;
      lfsr_next = 23'h7FFFFF;
    end

    sync_out = !acc[23] && acc_next[23];
  end
endmodule
