`default_nettype none

module sid_env (
  input wire clk,
  input wire clk_en,
  input wire n_reset,
  input wire gate,
  input wire [3:0] atk,
  input wire [3:0] dcy,
  input wire [3:0] stn,
  input wire [3:0] rls,
  output reg [7:0] vol
  );

  localparam
    ATTACK        = 2'd0,
    DECAY_SUSTAIN = 2'd1,
    RELEASE       = 2'd2;

  reg [14:0] adsrtable[0:15];
  initial begin
    adsrtable[0]  = 15'h007F;
    adsrtable[1]  = 15'h3000;
    adsrtable[2]  = 15'h1E00;
    adsrtable[3]  = 15'h0660;
    adsrtable[4]  = 15'h0182;
    adsrtable[5]  = 15'h5573;
    adsrtable[6]  = 15'h000E;
    adsrtable[7]  = 15'h3805;
    adsrtable[8]  = 15'h2424;
    adsrtable[9]  = 15'h2220;
    adsrtable[10] = 15'h090C;
    adsrtable[11] = 15'h0ECD;
    adsrtable[12] = 15'h010E;
    adsrtable[13] = 15'h23F7;
    adsrtable[14] = 15'h5237;
    adsrtable[15] = 15'h64A8;
  end

  reg [1:0] state, state_next;

  reg [7:0]  vol_next;
  reg [14:0] lfsr, lfsr_next;

  reg [4:0] exp_counter, exp_counter_next;
  reg [4:0] exp_period, exp_period_next;

  always @(posedge clk) begin
    if (!n_reset) begin
      state <= RELEASE;
      vol <= 0;
      lfsr <= 15'h7FFF;
      exp_counter <= 0;
      exp_period  <= 1;
    end
    else if (clk_en) begin
      state <= state_next;
      vol <= vol_next;
      lfsr <= lfsr_next;
      exp_counter <= exp_counter_next;
      exp_period  <= exp_period_next;
    end
  end

  always @* begin
    state_next = state;
    vol_next = vol;
    exp_period_next = exp_period;

    case (vol)
      'hFF:   exp_period_next = 1;
      'h5D:   exp_period_next = 2;
      'h36:   exp_period_next = 4;
      'h1A:   exp_period_next = 8;
      'h0E:   exp_period_next = 16;
      'h06:   exp_period_next = 30;
      'h00:   exp_period_next = 1;
    endcase

    if (exp_counter == exp_period) begin
      exp_counter_next = 0;
    end
    else begin
      exp_counter_next = exp_counter + 1;
    end

    if (exp_counter == 0 || state == ATTACK)
      lfsr_next = { lfsr[1] ^ lfsr[0], lfsr[14:1] };
    else
      lfsr_next = lfsr;

    case (state)  // Esto debera implementarse como un parallel case
      ATTACK:
        begin
          if (lfsr == adsrtable[atk]) begin
            lfsr_next = 15'h7FFF;
            if (vol != 255)
              vol_next = vol + 1;
            else
              state_next = DECAY_SUSTAIN;
          end

          if (!gate)
            state_next = RELEASE;
        end

      DECAY_SUSTAIN:
        begin
          if (lfsr == adsrtable[dcy]) begin
            lfsr_next = 15'h7FFF;
            if (vol != { stn, stn })
              vol_next = vol - 1;
          end

          if (!gate)
            state_next = RELEASE;
        end

        RELEASE:
          begin
            if (lfsr == adsrtable[rls]) begin
              lfsr_next = 15'h7FFF;
              if (vol != 0)
                vol_next = vol - 1;
            end

            if (gate)
              state_next = ATTACK;
          end

        default:
          state_next = RELEASE;
    endcase
  end
endmodule

