`default_nettype none

module sid_filter (
  input wire           clk,
  input wire           clk_en,
  input wire           n_reset,
  input wire [11:0]    v0,
  input wire [11:0]    v1,
  input wire [11:0]    v2,
  input wire [10:0]    reg_fc,
  input wire [3:0]     reg_res,
  input wire [3:0]     reg_en,
  input wire           reg_off3,
  input wire           reg_hp,
  input wire           reg_bp,
  input wire           reg_lp,
  input wire [3:0]     reg_vol,
  output reg [15:0]    audio_out
  );

  reg signed [31:0] low,  low_next;
  reg signed [31:0] band, band_next;
  reg signed [31:0] high, high_next;
  reg signed [31:0] out_next;

  always @(posedge clk) begin
    if (!n_reset) begin
      low  <= 0;
      band <= 0;
      high <= 0;
      audio_out <= 0;
    end
    else if (clk_en) begin
      low  <= low_next;
      band <= band_next;
      high <= high_next;
      audio_out <= out_next;
    end
  end

  reg signed [31:0] out_filt;
  reg signed [31:0] fc, res;
  always @* begin
    // 3*12 Bit Voice -> 14 bit out_filt/next
    //
    out_filt = 0;
    out_next = 0;

    if (reg_en[0])
      out_filt = out_filt + v0;
    else
      out_next = out_next + v0;

    if (reg_en[1])
      out_filt = out_filt + v1;
    else
      out_next = out_next + v1;

    if (reg_en[2])
      out_filt = out_filt + v2;
    else if (!reg_off3)
      out_next = out_next + v2;

    fc  = reg_fc + 64;
    res = 256 - reg_res * 10;

    high_next = out_filt - low - ((band * res) >>> 8);
    band_next = band + ((high_next * fc) >>> 16);
    low_next  = low  + ((band_next * fc) >>> 16);

    if (reg_lp)  out_next += low;
    if (reg_bp)  out_next += band;
    if (reg_hp)  out_next += high;

    out_next = (out_next * $signed({1'b0,reg_vol})) >>> 2;

    if (out_next < 0)
        out_next = 0;
    if (out_next > 65535)
        out_next = 65535;
  end
endmodule
