`default_nettype none

module sid_wave (
  input wire [23:0] acc,
  input wire [22:0] lfsr,
  input wire [7:0] vol,
  input wire noise, 
  input wire pulse, 
  input wire saw, 
  input wire triangle,
  input wire [11:0] pw,
  input wire ring,
  input wire ring_in,
  output reg [11:0] out
  );

  reg [11:0] out_noise, out_pulse, out_saw, out_triangle;
  reg [19:0] outtemp;
  always @* begin
    out_noise = {
        lfsr[20], lfsr[18], lfsr[14], lfsr[11],
        lfsr[ 9], lfsr[ 5], lfsr[ 2], lfsr[ 0], 4'b0
    };

    out_pulse    = (acc[23:12] >= pw)? 12'hFFF : 12'h000;
    out_saw      = acc[23:12];
    out_triangle = {acc[22:12], 1'b0};

    if (ring)
       out_triangle = {12{ring_in}} ^ out_triangle;
    else
       out_triangle = {12{acc[23]}} ^ out_triangle;

    out = 12'hFFF;
    if (noise)      out = out & out_noise;
    if (pulse)      out = out & out_pulse;
    if (saw)        out = out & out_saw;
    if (triangle)   out = out & out_triangle;

    outtemp = {8'h00,out} * {12'h000,vol};
    out  = outtemp[19:8];
  end
endmodule