`timescale 1ns / 1ns

module tb_sid;
  reg clk;
  reg n_reset;
  wire [15:0] audio_out;
  reg [4:0] addr;
  reg [7:0] data;
  reg n_cs;
  reg rw;

  mos6581 sid (            // se instancia el chip
    .clk(clk),
    .clk_en(1'b1),
    .n_reset(n_reset),
    .addr(addr),
    .n_cs(n_cs),
    .rw(rw),
    .din(data),
    .audio_out(audio_out)
  );

  integer fdata, faudio;     // los dos manejadores de fichero
  integer res;               // variable que se usa en fputc
  integer frames;            // cuantos frames llevamos reproducidos
  initial begin
    //$dumpfile ("dump.vcd");  //
    //$dumpvars (0, tb_sid);   // Interfaz con GTKWave

    n_cs = 1;                // chip inactivo al principio
    rw = 1;
    clk = 0;
    n_reset = 0;             // y reset activo
    addr = 0;                // direccion del MOS 6581 donde se guardar el dato leido
    frames = 0;

    fdata = $fopen ("CybernoidII.bin", "rb");  // de aqu leemos el volcado de registros
    if (fdata == 0) begin
      $display ("Fichero de entrada no encontrado");
      $finish;
    end

    faudio = $fopen ("CybernoidII.raw", "wb");  // aqu escribiremos las muestras de audio. Con sox las traduciremos a un WAV
    #5000;                   // todo listo. Esperamos 5 microsegundos...
    n_reset = 1;             // y quitamos el reset
    n_cs = 0;                // habilitamos el chip
    res = $fscanf (fdata, "%c", data);  // llemos el primer dato
    while (!$feof(fdata)) begin
      @(posedge clk);
      rw = 0;                // pulsamos la escritura
      @(posedge clk);        // esperamos
      @(posedge clk);        // un poquito
      rw = 1;                // y desactivamos la escritura
      @(posedge clk);        // esperamos otro poco
      addr = addr + 1;       // siguiente direccion
      if (addr == 25) begin  // si ya hemos escrito 25 direcciones...
        addr = 0;               // volvemos a empezar desde la direccion 0
        frames = frames + 1;    // contamos 1 frame ms
        $display ("Frames: %0d", frames);
        #20000000;              // y esperamos 20 milisegundos
      end
      res = $fscanf (fdata, "%c", data);  // leemos el siguiente dato
    end
    $fclose (fdata);
    $fclose (faudio);
    $finish;
  end

  // este always se activa 44100 veces por segundo. Su misin es 
  // muestrear la salida de audio y guardar la muestra en un fichero
  always begin
    res = $fputc (audio_out[7:0], faudio);   // guardamos 16 bits en formato little endian. Primero los 8 bits menos significativos
    res = $fputc (audio_out[15:8], faudio);  // y luego los 8 bits ms significativos
    #(1000000000.0 / 44100);                 // esperar un ciclo de reloj de 44.1 kHz
  end

  // este es el reloj de 1 MHz para el SID
  always begin
    clk = #500 ~clk;
  end
endmodule